/* Telefunken RT 200 Firmware
 * (c) 9.10.2016 Ingo D. Rullhusen
 */

/* Simple example for Teensy USB Development Board
 * http://www.pjrc.com/teensy/
 * Copyright (c) 2008 PJRC.COM, LLC
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <stdint.h>
#include <util/delay.h>

#include <stdio.h>
#include <string.h>

#include "usb_serial.h"

#include "diskey.h"
#include "timer.h"
#include "tuner.h"


void wdt_init(void) __attribute__((naked)) __attribute__((section(".init3")));

void wdt_init(void)
{
    MCUSR = 0;
    wdt_disable();

    return;
}


void send_str(const char *s);
uint8_t recv_str(char *buf, uint8_t size);
void parse_and_execute_command(const char *buf, uint8_t num);


void recv_str_noblock(char *buf, uint8_t *count, uint8_t size)
{
   int16_t r;

   while (*count < size)
   {
      r = usb_serial_getchar();
      if (r != -1)
      {
         if ( (r >= ' ' && r <= '~') || (r == '\r') || (r == '\n') )
         {
            usb_serial_putchar(r);
            buf[*count] = r;
            ++(*count);
         }
         if ( r == 127 )
         {
            if ( *count > 0 ) { send_str(PSTR("\b \b")); --(*count); }
         }
      }
      else
      {
         break;
      }
   }

}

void send_banner(void)
{
   send_str(PSTR(
      "\r\nTELEFUNKEN RT 200,\r\n"
      "HiFi-Tuner USB-serial interface\r\n\r\n"
      "Commands:\r\n"
      "  ?  this text\r\n"
      "  CALL  call user function\r\n"
      "  LED=[01?]  teensy LED\r\n"
      "  STORELED=[01?]  store key LED\r\n"
      "  LOCKDET?  lockdet pin\r\n"
      "  STATDET?  statdet pin\r\n"
      "  PWRDET?   pwrdet pin\r\n"
      "  KM?  keyboard matrix\r\n"
      "  KEYS?  key state\r\n"
      "  SEG=[00-ff]: ... :[00-ff]  set segmets\r\n"
      "  MUTE=[01?]   mute pin\r\n"
      "  POWER=[01?]  power pin\r\n"
      "  BCDLED=[0-f?]  preset LEDs\r\n"
      "  TIME=([0-f][0-f]:[0-f][0-f])|?  set/show time\r\n"
      "  BAND=[0124?]  band select pins\r\n"
      "  MWLO=<integer>|?  lower MW frequency in kHz\r\n"
      "  MWHI=<integer>|?  upper MW frequency in kHz\r\n"
      "  FMLO=<integer>|?  lower FM frequency in 10kHz\r\n"
      "  FMHI=<integer>|?  upper FM frequency in 10kHz\r\n"
      "  PLL=[0-f]: ... :[0-f]  set PLL register\r\n"
                ));
}

#define BUFFSIZE 32
void handle_usb(void)
{
   static char buf[BUFFSIZE];
   static uint8_t n=0, serial_state=0;

   if (serial_state)
   {
      if ( usb_configured() && (usb_serial_get_control() & USB_SERIAL_DTR) )
      {
         recv_str_noblock(buf, &n, sizeof(buf));
         if ( (n>0 && (buf[n-1] == '\r' || buf[n-1] == '\n')) || n == BUFFSIZE )
         {
            send_str(PSTR("\r\n"));
            parse_and_execute_command(buf, n);
            send_str(PSTR("\r\n>"));
            n = 0;
         }
      }
      else
      {
         serial_state = 0;
         LED_OFF;
      }
   }
   else
   {
      if ( usb_configured() && (usb_serial_get_control() & USB_SERIAL_DTR) )
      {
         LED_ON;
         serial_state = 1; n = 0;
         usb_serial_flush_input();
         send_banner();
         send_str(PSTR("\r\n>"));
      }
   }

}


// MAIN
int main(void)
{
   bcd_time lastClock;

// set for 4 MHz clock, and turn on the LED
   CPU_PRESCALE(2);
   LED_CONFIG;
   LED_OFF;

   while ( ! PWRDET() ) {}

// power off unused
   power_adc_disable();
   power_spi_disable();
   power_twi_disable();
   power_timer1_disable();
   power_usart1_disable();

   usb_init();

   diskey_init();
   timer_init();
   tuner_init();

// force init
   while ( CLOCK.H > 2 )
   {
      diskey_str2seg( PSTR("E    "), 1, 0 );

      if ( MODE != M_BATPWR ) { handle_usb(); }

      if ( MODE == M_CSET )
      {
         bcd_time t_t;
         if ( time_input( &t_t ) )
         {
            while ( MODE == M_CSET ) {}
            time_set(&t_t);
         }
      }

      set_sleep_mode(SLEEP_MODE_IDLE);
      cli();
      sleep_enable();
      sei();
      sleep_cpu();
      sleep_disable();
   }

// main
   lastClock = CLOCK;
   while (1)
   {
      if ( MODE != M_BATPWR ) { handle_usb(); }

      switch ( MODE )
      {
         case M_ON:
         {
            if ( (lastClock.m != CLOCK.m)
                 || (lastClock.M != CLOCK.M)
                    || (lastClock.h != CLOCK.h)
                       || (lastClock.H != CLOCK.H) )
            {
               uint8_t intr_state = SREG;
               cli();
               lastClock = CLOCK;
               SREG = intr_state;
               
               if ( tuner_pwr_state() )
               {
                  if      ( (lastClock.m == ONCE.off.m)
                            && (lastClock.M == ONCE.off.M)
                               && (lastClock.h == ONCE.off.h)
                                  && (lastClock.H == ONCE.off.H) )
                  {
                     tuner_pwr_off();
                     ONCE.off.H = ONCE.off.h = ONCE.off.M = ONCE.off.m = 0xff;
                  }
                  else if ( (lastClock.m == PERMANENT.off.m)
                            && (lastClock.M == PERMANENT.off.M)
                               && (lastClock.h == PERMANENT.off.h)
                                  && (lastClock.H == PERMANENT.off.H) )
                  {
                     tuner_pwr_off();
                  }
               }
               else
               {
                  if      ( (lastClock.m == ONCE.on.m)
                            && (lastClock.M == ONCE.on.M)
                               && (lastClock.h == ONCE.on.h)
                                  && (lastClock.H == ONCE.on.H) )
                  {
                     last_band = ONCE.band;
                     if ( last_band == BAND_MW )
                     {
                        last_am_freq = ( ONCE.station < 8
                                         ? AMstation[ ONCE.station ]
                                         : 0 );
                        last_am_pLED = ONCE.station;
                     }
                     else
                     {
                        last_fm_freq = ( ONCE.station < 8
                                         ? FMstation[ ONCE.station ]
                                         : 0 );
                        last_fm_pLED = ONCE.station;
                     }
                     tuner_pwr_on();
                     ONCE.on.H = ONCE.on.h = ONCE.on.M = ONCE.on.m = 0xff;
                     ONCE.band = BAND_OFF;
                     ONCE.station = 0xff;
                  }
                  else if ( (lastClock.m == PERMANENT.on.m)
                            && (lastClock.M == PERMANENT.on.M)
                               && (lastClock.h == PERMANENT.on.h)
                                  && (lastClock.H == PERMANENT.on.H) )
                  {
                     last_band = PERMANENT.band;
                     if ( last_band == BAND_MW )
                     {
                        last_am_freq = ( PERMANENT.station < 8
                                         ? AMstation[ PERMANENT.station ]
                                         : 0 );
                        last_am_pLED = PERMANENT.station;
                     }
                     else
                     {
                        last_fm_freq = ( PERMANENT.station < 8
                                         ? FMstation[ PERMANENT.station ]
                                         : 0 );
                        last_fm_pLED = PERMANENT.station;
                     }
                     tuner_pwr_on();
                  }
               }
            }
            /* no break */
         }
         case M_OFF:
         {
            if ( tuner_pwr_state() )
            { tuner_on(); }
            else
            { tuner_wait(); }
            break;
         }
         case M_TSET:
         {
            tuner_pwr_off();
            timer_set();
            break; 
         }
         case M_CANCEL:
         {
            timer_cancel();
            break; 
         }
         case M_CHECK:
         {
            tuner_pwr_off();
            timer_check();
            break;
         }
         case M_CSET:
         {
            bcd_time t_t;
            if ( time_input( &t_t ) )
            {
               while ( MODE == M_CSET ) {}
               time_set(&t_t);
            }
            break;
         }
         case M_BATPWR:
         { break; }

         case M_RESET:
         {
            diskey_str2seg( PSTR("rEsEt"), 0, 0 );
            if ( KEY == K_RESET )
            {
               for ( int i=0; i<30 && KEY == K_RESET && MODE == M_RESET; ++i )
               { _delay_ms(100); }
               if ( KEY == K_RESET )
               {
                  wdt_enable(WDTO_15MS);
                  for(;;) {}
               }
            }
            break;
         }
         case M_MSG:
         {
            show_message();
            break;
         }

         default: /* error */
         {
            uint8_t mode_t = MODE;
            for ( int i=0; i<30 && MODE == mode_t; ++i )
            { _delay_ms(100); }
            if ( MODE == mode_t )
            { diskey_str2seg( PSTR("Error"), 0, 0 ); }
         }
      }

      if ( tuner_pwr_state() && KEY == K_PWROFF )
      {
         tuner_pwr_off();
         diskey_str2seg( PSTR(" byE "), 0, 0 );
         _delay_ms(1500);
      }

      set_sleep_mode(SLEEP_MODE_IDLE);
      cli();
      sleep_enable();
      sei();
      sleep_cpu();
      sleep_disable();

//      PINA |= 0x40;

   }
}


// Send a string to the USB serial port.  The string must be in
// flash memory, using PSTR
//
void send_str(const char *s)
{
	char c;
	while (1) {
		c = pgm_read_byte(s++);
		if (!c) break;
		usb_serial_putchar(c);
	}
}

// Receive a string from the USB serial port.  The string is stored
// in the buffer and this function will not exceed the buffer size.
// A carriage return or newline completes the string, and is not
// stored into the buffer.
// The return value is the number of characters received, or 255 if
// the virtual serial connection was closed while waiting.
//
uint8_t recv_str(char *buf, uint8_t size)
{
	int16_t r;
	uint8_t count=0;

	while (count < size) {
		r = usb_serial_getchar();
		if (r != -1) {
			if (r == '\r' || r == '\n') return count;
			if (r >= ' ' && r <= '~') {
				*buf++ = r;
				usb_serial_putchar(r);
				count++;
			}
		} else {
			if (!usb_configured() ||
			  !(usb_serial_get_control() & USB_SERIAL_DTR)) {
				// user no longer connected
				return 255;
			}
			// just a normal timeout, keep waiting
		}
	}
	return count;
}

// parse a user command and execute it, or print an error message
//
void parse_and_execute_command(const char *buf, uint8_t num)
{
   char obuff[BUFFSIZE];
   uint8_t val;
   unsigned int ui;

   if ( buf[0] == '\r' || buf[0] == '\n' ) { return; }

   if ( strncmp("?",buf,1) == 0 )
   {
      send_banner();
      return;
   }

   if ( strncmp("CALL",buf,4) == 0 )
   {

      send_str(PSTR("NOP\r\n"));

      return;
   }

   if ( strncmp("LED=",buf,4) == 0 )
   {
      if ( buf[4] == '0' ) { LED_OFF; return; }
      if ( buf[4] == '1' ) { LED_ON; return; }
      if ( buf[4] == '?' )
      {
         if ( PORTD & 0x80 ) { usb_serial_putchar('1'); }
         else                { usb_serial_putchar('0'); }
         send_str(PSTR("\r\n"));
         return;
      }
   }

   if ( strncmp("STORELED=",buf,9) == 0 )
   {
      if ( buf[9] == '0' ) { diskey_sLED(0); return; }
      if ( buf[9] == '1' ) { diskey_sLED(1); return; }
      if ( buf[9] == '?' )
      {
         if ( SEGMENT[0] & 0x20 ) { usb_serial_putchar('0'); }
         else                     { usb_serial_putchar('1'); }
         send_str(PSTR("\r\n"));
         return;
      }
   }

   if ( strncmp("LOCKDET?",buf,8) == 0 )
   {
      if ( LOCKDET() ) { usb_serial_putchar('1'); }
      else             { usb_serial_putchar('0'); }
      send_str(PSTR("\r\n"));
      return;
   }

   if ( strncmp("STATDET?",buf,8) == 0 )
   {
      if ( STATDET() ) { usb_serial_putchar('1'); }
      else             { usb_serial_putchar('0'); }
      send_str(PSTR("\r\n"));
      return;
   }

   if ( strncmp("PWRDET?",buf,7) == 0 )
   {
      if ( PWRDET() ) { usb_serial_putchar('1'); }
      else            { usb_serial_putchar('0'); }
      send_str(PSTR("\r\n"));
      return;
   }

   if ( strncmp("KM?",buf,3) == 0 )
   {
      for ( int i=0; i<6; ++i )
      {
         snprintf(obuff,BUFFSIZE,"%.2x",key_matrix[i]);
         usb_serial_write((uint8_t*)obuff,2);
         if ( i<5 ) { send_str(PSTR(":")); }
      }
      send_str(PSTR("\r\n"));
      return;
   }

   if ( strncmp("KEYS?",buf,5) == 0 )
   {
      if ( AUTOSCAN ) { send_str(PSTR("AUTOSCAN")); }
      else            { send_str(PSTR("-")); }
      usb_serial_putchar(':');
      switch( MODE )
      {
         case M_ON: { send_str(PSTR("TIMER ON")); break; }
         case M_OFF: { send_str(PSTR("TIMER OFF")); break; }
         case M_TSET: { send_str(PSTR("TIMER SET")); break; }
         case M_CANCEL: { send_str(PSTR("TIMER CANCEL")); break; }
         case M_CHECK: { send_str(PSTR("CHECK")); break; }
         case M_CSET: { send_str(PSTR("CLOCK SET")); break; }
         case M_BATPWR: { send_str(PSTR("battery")); break; }
         default: { send_str(PSTR("-")); break; }
      }
      usb_serial_putchar(':');
      switch( KEY )
      {
         case K_0_MW: { send_str(PSTR("0/MW")); break; }
         case K_1: { send_str(PSTR("1")); break; }
         case K_2: { send_str(PSTR("2")); break; }
         case K_3: { send_str(PSTR("3")); break; }
         case K_4: { send_str(PSTR("4")); break; }
         case K_5: { send_str(PSTR("5")); break; }
         case K_6: { send_str(PSTR("6")); break; }
         case K_7: { send_str(PSTR("7")); break; }
         case K_8: { send_str(PSTR("8")); break; }
         case K_9_FM: { send_str(PSTR("9/FM")); break; }
         case K_STORE: { send_str(PSTR("STORE")); break; }
         case K_LEFT: { send_str(PSTR("<")); break; }
         case K_RIGHT: { send_str(PSTR(">")); break; }
         case K_PWROFF: { send_str(PSTR("OFF")); break; }
         default: { send_str(PSTR("-")); break; }
      }
      send_str(PSTR("\r\n"));
      return;
   }

   if ( strncmp("MUTE=",buf,5) == 0 )
   {
      if ( buf[5] == '0' ) { set_mute(0); return; }
      if ( buf[5] == '1' ) { set_mute(1); return; }
      if ( buf[5] == '?' )
      {
         if ( PORTE & 0x02 ) { usb_serial_putchar('1'); }
         else                { usb_serial_putchar('0'); }
         send_str(PSTR("\r\n"));
         return;
      }
   }

   if ( strncmp("POWER=",buf,6) == 0 )
   {
      if ( buf[6] == '0' ) { set_power(0); return; }
      if ( buf[6] == '1' ) { set_power(1); return; }
      if ( buf[6] == '?' )
      {
         if ( PORTE & 0x01 ) { usb_serial_putchar('1'); }
         else                { usb_serial_putchar('0'); }
         send_str(PSTR("\r\n"));
         return;
      }
   }

   if ( strncmp("BAND=",buf,5) == 0 )
   {
      switch (buf[5])
      {
         case '0': { set_band(BAND_OFF); return; }
         case '1': { set_band(BAND_LW ); return; }
         case '2': { set_band(BAND_MW ); return; }
         case '4': { set_band(BAND_FM ); return; }
         case '?':
         {
            val = PORTC & 0x07;
            snprintf(obuff,BUFFSIZE,"%1x",val);
            usb_serial_putchar(obuff[0]);
            send_str(PSTR("\r\n"));
            return;
         }
      }
   }

   if ( strncmp("BCDLED=",buf,7) == 0 )
   {
      if ( buf[7] == '?' )
      {
         val = (PORTC & 0xf0)>>4;
         snprintf(obuff,BUFFSIZE,"%1x",val);
         usb_serial_putchar(obuff[0]);
         send_str(PSTR("\r\n"));
         return;
      }

      sscanf(buf+7,"%x", &ui );
      diskey_pLED( (uint8_t)ui );
      return;

   }

   if ( strncmp("TIME=",buf,5) == 0 )
   {
      bcd_time clk;

      if ( buf[5] == '?' )
      {
         snprintf(obuff,BUFFSIZE,"%.1x", CLOCK.H );
         usb_serial_putchar(obuff[0]);
         snprintf(obuff,BUFFSIZE,"%.1x", CLOCK.h );
         usb_serial_putchar(obuff[0]);
         send_str(PSTR(":"));
         snprintf(obuff,BUFFSIZE,"%.1x", CLOCK.M );
         usb_serial_putchar(obuff[0]);
         snprintf(obuff,BUFFSIZE,"%.1x", CLOCK.m );
         usb_serial_putchar(obuff[0]);

         ui = TCNT3;
         snprintf(obuff,BUFFSIZE,":%.5u/15000", ui );
         usb_serial_write((uint8_t*)obuff,12);

         send_str(PSTR("\r\n"));
         return;
      }

      sscanf(buf+5,"%1x", &ui );
      clk.H = ui;
      sscanf(buf+6,"%1x", &ui );
      clk.h = ui;
      sscanf(buf+8,"%1x", &ui );
      clk.M = ui;
      sscanf(buf+9,"%1x", &ui );
      clk.m = ui;
      time_set( &clk );

      return;
   }

   if ( strncmp("SEG=",buf,4) == 0 )
   {
      for ( int i=0; i<6; ++i )
      {
         sscanf(buf+4+3*i,"%x", &ui );
         SEGMENT[i] = ui;
      }
      return;
   }

   if ( strncmp("MWLO=",buf,5) == 0 )
   {
      if ( buf[5] == '?' )
      {
         snprintf(obuff,BUFFSIZE,"%i",am_lower);
         usb_serial_write((uint8_t*)obuff,strnlen(obuff,BUFFSIZE));
         send_str(PSTR("\r\n"));
         return;
      }

      sscanf(buf+5,"%i", &am_lower );

      return;
   }

   if ( strncmp("MWHI=",buf,5) == 0 )
   {
      if ( buf[5] == '?' )
      {
         snprintf(obuff,BUFFSIZE,"%i",am_upper);
         usb_serial_write((uint8_t*)obuff,strnlen(obuff,BUFFSIZE));
         send_str(PSTR("\r\n"));
         return;
      }

      sscanf(buf+5,"%i", &am_upper );

      return;
   }

   if ( strncmp("FMLO=",buf,5) == 0 )
   {
      if ( buf[5] == '?' )
      {
         snprintf(obuff,BUFFSIZE,"%i",fm_lower);
         usb_serial_write((uint8_t*)obuff,strnlen(obuff,BUFFSIZE));
         send_str(PSTR("\r\n"));
         return;
      }

      sscanf(buf+5,"%i", &fm_lower );

      return;
   }

   if ( strncmp("FMHI=",buf,5) == 0 )
   {
      if ( buf[5] == '?' )
      {
         snprintf(obuff,BUFFSIZE,"%i",fm_upper);
         usb_serial_write((uint8_t*)obuff,strnlen(obuff,BUFFSIZE));
         send_str(PSTR("\r\n"));
         return;
      }

      sscanf(buf+5,"%i", &fm_upper );

      return;
   }

   if ( strncmp("PLL=",buf,4) == 0 )
   {
      if ( buf[4] == '?' )
      {
         if ( diskey_set_pll ) { usb_serial_putchar('1'); }
         else                  { usb_serial_putchar('0'); }
         send_str(PSTR("\r\n"));
         return;
      }

      for ( int i=0; i<8; ++i )
      {
         sscanf(buf+4+2*i,"%1x", &ui );
         tuneFreqR[i] = ui;
      }
      diskey_set_pll = PLL_SET;

      return;
   }

// otherwise, error message
   send_str(PSTR("Unknown command.\r\n"));
}

